(*----------------------------------------------------------------------------*
 *  Direct3D DepthColoring sample                                             *
 *  demostrate how to convert depth information to color during vertex        *
 *  processing. DirectX 8.1 headers and lib's are required.                   *
 *  Above mentioned effect is generated using 3 different methods:            *
 *  1) Good old texture matrix transformation using DX7 style fixed function  *
 *     T&L and accessing small pre-rendered texture while pixel shading.      *
 *  2) Modern style DX8 VertexShader version of first method - 40% faster on  *
 *     GF3 on complex meshes. Also accesses prerendered textule later.        *
 *  3) Calculating color (hue channel of HSL color model) completely in       *
 *     vertex shader. Pros: no texture needed - ouputs standart diffuse color;*
 *     Cons: more complex calculations required - less vertex processing speed*
 *  So, recommendations: on hardware capable of VS use (2) else (1), if you   *
 *  need many textures when (3) may be win for your setup.                    *
 *                                                                            *
 *  Note: This sample requires DirectX 8.1 compatible headers to compile !!!  *
 *                                                                            *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 14-Sep-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

{$I DirectX.inc}

unit DepthColoringCGUnit;

interface

uses
  Windows, Messages, SysUtils, Math, CommDlg,
  D3DX8, {$I UseD3D8.inc}, D3DApp, D3DFont, D3DUtil, DXUtil, DXFile, D3DFile,
  Resource, cgD3D;

const
  SampleTitle = 'DepthColoring CG';

var
  cg_: TcgDirect3D;

  pVertexProgramContainer : TcgProgramContainer = nil;

  vertex_tot_iter         : PcgBindIter = nil;
  vertex_matWVColor_iter  : PcgBindIter = nil;

  pContextContainer: TcgContextContainer = nil;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_strMeshFilename:  array [0..511] of Char;
    m_strInitialDir:    array [0..511] of Char;

    m_pObject: CD3DMesh;                    // Object to render
    m_fSpectrumRange:    Single;

    m_bShowHelp:        Boolean;
    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;
    m_ArcBall:          CD3DArcBall;        // mouse rotation utility
    m_vObjectCenter:    TD3DXVector3;       // Center of bounding sphere of object
    m_fObjectRadius:    Single;             // Radius of bounding sphere of object

    matProj, matTotal, matWorldView: TD3DXMatrix;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    destructor Destroy; override;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
 end;

implementation

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);

  m_strWindowTitle    := SampleTitle;
  m_bUseDepthBuffer   := True;
  m_bShowCursorWhenFullscreen := True;

  m_pObject           := CD3DMesh.Create;
  m_pFont             := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall        := CD3DFont.Create('Arial', 10, D3DFONT_BOLD);
  m_ArcBall           := CD3DArcBall.Create;
  m_bShowHelp         := False;

  StrCopy(m_strInitialDir, DXUtil_GetDXSDKMediaPath);
  m_strMeshFilename:= 'tiger.x';

  m_fSpectrumRange:= 1;
end;

destructor CMyD3DApplication.Destroy;
begin
  m_ArcBall.Free;
  inherited;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  cg_:= TcgDirect3D.Create;

  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  matWorld: TD3DXMatrix;
  matView: TD3DXMatrix;
begin
   // Setup world matrix
  D3DXMatrixTranslation(matWorld, -m_vObjectCenter.x,
                                  -m_vObjectCenter.y,
                                  -m_vObjectCenter.z);
  D3DXMatrixMultiply(matWorld, matWorld, m_ArcBall.GetRotationMatrix^);
  D3DXMatrixMultiply(matWorld, matWorld, m_ArcBall.GetTranslationMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);

  D3DXMatrixLookAtLH(matView, D3DXVector3(0, 0, -2*m_fObjectRadius),
                              D3DXVector3(0, 0, 0),
                              D3DXVector3(0, 1, 0));
  m_pd3dDevice.SetTransform(D3DTS_VIEW,  matView);

  // Calculate resulting matrix
  D3DXMatrixMultiply(matWorldView, matWorld, matView);
  D3DXMatrixMultiply(matTotal, matWorldView, matProj);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
var
  m: TD3DXMatrix;

  vb: IDirect3DVertexBuffer8;
  ib: IDirect3DIndexBuffer8;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $00FF, 1, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Activate shader
    pVertexProgramContainer.SetShaderActive;

    // Fill VS parameters 1
    pVertexProgramContainer.SetShaderConstant(vertex_tot_iter, matTotal);
    // Fill VS parameters 2
    FillChar(m, SizeOf(m), 0);
    m._31:= m_fSpectrumRange/m_fObjectRadius;
    D3DXMatrixMultiply(m, matWorldView, m);
    pVertexProgramContainer.SetShaderConstant(vertex_matWVColor_iter, m);

    // Acquire mesh data
    m_pObject.m_pLocalMesh.GetVertexBuffer(vb);
    m_pObject.m_pLocalMesh.GetIndexBuffer(ib);

    m_pd3dDevice.SetStreamSource(0, vb, 3*4);
    m_pd3dDevice.SetIndices(ib, 0);


    // Draw mesh
    m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0,
      m_pObject.m_pLocalMesh.GetNumVertices, 0, m_pObject.m_pLocalMesh.GetNumFaces);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,200,200,100),
                            'Keyboard controls:' );
      m_pFontSmall.DrawText(20, 60, D3DCOLOR_ARGB(255,200,200,100),
                            'Next Technique'#10 +
                            'Next Tech. (no validate)'#10 +
                            'Prev Technique'#10 +
                            'Prev Tech. (no validate)'#10 +
                            'Shrink range'#10'Expand range'#10 +
                            'Help'#10'Change device'#10'Exit');
      // Draw controls with high contrast
      m_pFontSmall.DrawText(181, 61, D3DCOLOR_ARGB(255,0,0,0),
                            'PageDn'#10'Shift-PageDn'#10 +
                            'PageUp'#10'Shift-PageUp'#10 +
                            'Gray +'#10'Gray -'#10 +
                            'F1'#10'F2'#10'Esc');
      m_pFontSmall.DrawText(180, 60, D3DCOLOR_ARGB(255,255,255,255),
                            'PageDn'#10'Shift-PageDn'#10 +
                            'PageUp'#10'Shift-PageUp'#10 +
                            'Gray +'#10'Gray -'#10 +
                            'F1'#10'F2'#10'Esc');
    end else
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,200,200,100),
                            'Press F1 for help' );
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: This creates all device-dependent managed objects, such as managed
//       textures and managed vertex buffers.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  pVB: IDirect3DVertexBuffer8;
  pVertices: PByte;
  pTempMesh: ID3DXMesh;
  strMeshPath: array[0..511] of Char;
begin
  pVertices := nil;

  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  // Load the mesh from the specified file
  DXUtil_FindMediaFile(strMeshPath, m_strMeshFilename);

  // Load an object to render
  Result:= m_pObject.Create_(m_pd3dDevice, m_strMeshFilename);
  if FAILED(Result) then
  begin
    // ignore load errors, just draw blank screen if mesh is invalid
    Result:= D3DAPPERR_MEDIANOTFOUND; // S_OK;
    Exit;
  end;

  // Convert to simple mesh need for sample
  with m_pObject.m_pSysMemMesh do
    Result:= CloneMeshFVF(GetOptions, D3DFVF_XYZ, m_pd3dDevice, pTempMesh);
  if FAILED(Result) then Exit;

  m_pObject.m_pSysMemMesh:= pTempMesh;
  pTempMesh:= nil;

  // Lock the vertex buffer, to generate a simple bounding sphere
  Result:= m_pObject.m_pSysMemMesh.GetVertexBuffer(pVB);
  if FAILED(Result) then Exit;

  Result:= pVB.Lock(0, 0, pVertices, 0);
  if FAILED(Result) then Exit; // VB will be released by Delphi compiler

  Result:= D3DXComputeBoundingSphere(pVertices^, m_pObject.m_pSysMemMesh.GetNumVertices,
                                     m_pObject.m_pSysMemMesh.GetFVF, m_vObjectCenter,
                                     m_fObjectRadius);
  pVB.Unlock;
  pVB:= nil;

  if FAILED(Result) then Exit;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  light: TD3DLight8;
  m: TD3DXMatrix;
  fAspect: Single;
  listing, object_code, vert: PChar;
begin
  // Setup render state
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,      $33333333);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);

  // Setup the light
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.0,-1.0, 1.0);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, TRUE);

  // Setup the arcball parameters
  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.85);
  m_ArcBall.SetRadius(1.0);

  // Setup the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect,
                             m_fObjectRadius/64.0, m_fObjectRadius*200.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Restore the font
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  m_pObject.RestoreDeviceObjects(m_pd3dDevice);

  m_pd3dDevice.SetVertexShader(D3DFVF_XYZ);

  //cg.AddFilePath("C:\\AddYourCGDirectoryHere");
  cg_.AddFilePath('..');

  pContextContainer := cg_.CreateContextContainer(m_pd3dDevice);

  // shows how to use different vertex declaration than cg specifies
  pVertexProgramContainer := pContextContainer.LoadCGProgramFromFile(
    'Depth.cg', 'Depth Coloring Vertex Shader', cgDX8VertexProfile);

  if (pVertexProgramContainer = nil) then
  begin
    // listing in case of an error
    listing := pContextContainer.GetLastListing;

    if (listing = nil) then
      listing := 'Could not find cgc.exe.';

    cg_.NotePad('Failed to Create Vertex Program'#10#10, listing);

    Halt(1);
  end else
  begin
    // vertex shader asm code that is generated
    object_code := pVertexProgramContainer.GetProgramObjectCode;

    // uncomment this to see the vertex code
    //cg_.NotePad('', object_code);


    vert := pVertexProgramContainer.GetVertexDeclaration;

    // this should be how your vertex is defined so it matches the cg
    // uncomment this to see how your cg program is expecting the vertex to be defined
    //cg_.NotePad('cg program expects the vertex be defined like this:'#10, vert);
  end;

  // the are the variables from the Depth.cg program
  vertex_tot_iter := pVertexProgramContainer.GetParameterBindByName('mTotal');
  vertex_matWVColor_iter:= pVertexProgramContainer.GetParameterBindByName('mWVColor');

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pObject.InvalidateDeviceObjects;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;
  pContextContainer.Destroy;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pFontSmall);
  FreeAndNil(m_pObject);
//  cg_.Free;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
{$WRITEABLECONST ON}
const
  ofn: TOpenFilename = (
    lStructSize: SizeOf(TOpenFilename);
    lpstrFilter: '.X Files (.x)'#0'*.x'#0#0; nFilterIndex: 1;
    nMaxFile: 512;
    nMaxFileTitle: 512;
    lpstrTitle: 'Open Mesh File';
    Flags: OFN_FILEMUSTEXIST;
    nFileExtension: 1
  );
{$WRITEABLECONST OFF}
var
  g_strFilename: array [0..511] of Char;
  pLastSlash: PChar;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMouseMessages(hWnd, uMsg, wParam, lParam);

  // Trap context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result := 0;
    Exit;
  end;

  // Handle key presses
  if (WM_KEYDOWN = uMsg) then
  begin
    if (Ord('W') = wParam)
    then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME);

    if (Ord('S') = wParam)
    then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

    if (VK_ADD = wParam) then m_fSpectrumRange:= m_fSpectrumRange * 1.01;
    if (VK_SUBTRACT = wParam) then m_fSpectrumRange:= m_fSpectrumRange * 1/1.01;
  end
  else if (uMsg = WM_COMMAND) then
  begin
    // Handle the open file command
    if (LOWORD(wParam) = IDM_OPENFILE) then
    begin
      // Display the OpenFileName dialog. Then, try to load the specified file
      g_strFilename:= '';

      {$IFDEF COMPILER6_UP}
      // Handle Delphi6 case of extended OFN structure
      // (new members are invalid for Win98SE & below OS'es)
      ofn.lStructSize:= 76;

      {$ENDIF}
      ofn.lpstrFile:= m_strMeshFilename;
      ofn.lpstrFileTitle:= g_strFilename;
      ofn.lpstrInitialDir:= m_strInitialDir;

      if GetOpenFileName(ofn) then
      begin
        StrCopy(m_strInitialDir, m_strMeshFilename);
        pLastSlash:= StrPos(m_strInitialDir, '\');
        if (pLastSlash <> nil) then pLastSlash^ := #0;

        SetCurrentDirectory(m_strInitialDir);

        // Destroy and recreate everything
        InvalidateDeviceObjects;
        DeleteDeviceObjects;
        InitDeviceObjects;
        RestoreDeviceObjects;
      end;

    end;

    case LOWORD(wParam) of
     IDM_TOGGLEHELP:                  m_bShowHelp:= not m_bShowHelp;
    end;
  end;

  // Pass remaining messages to default handler
  Result := inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord;
  Format: TD3DFormat): HResult;
begin
  if( (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0 then 
  begin
    if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

end.

